package org.jastadd.ragconnect.tests;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import static org.assertj.core.api.Assertions.assertThat;
import static org.jastadd.ragconnect.tests.TestUtils.readFile;
import static org.junit.jupiter.api.Assertions.assertTrue;

public class Errors {

  private static final Logger logger = LogManager.getLogger(Errors.class);
  private static final String FILENAME_PATTERN = "$FILENAME";
  private static final String ERROR_DIRECTORY = "errors/";
  private static final String OUTPUT_DIRECTORY = TestUtils.OUTPUT_DIRECTORY_PREFIX + ERROR_DIRECTORY;

  private static final String DEFAULT_GRAMMAR_NAME = "Errors";

  @BeforeAll
  public static void createOutputDirectory() {
    File outputDirectory = new File(OUTPUT_DIRECTORY);
    assertTrue((outputDirectory.exists() && outputDirectory.isDirectory()) || outputDirectory.mkdir());
  }

  @Test
  void testStandardErrors() throws IOException {
    test("Standard", "A","Standard");
  }

  @Test
  void testTwoPartsErrors() throws IOException {
    test("Part", "A","Part1", "Part2");
  }

  @SuppressWarnings("SameParameterValue")
  private void test(String expectedName, String rootNode, String... connectNames) throws IOException {
    String grammarFile = ERROR_DIRECTORY + DEFAULT_GRAMMAR_NAME + ".relast";
    List<String> connectFiles = Arrays.stream(connectNames)
        .map(connectName -> ERROR_DIRECTORY + connectName + ".connect")
        .collect(Collectors.toList());
    Path outPath = TestUtils.runCompiler(grammarFile, connectFiles, rootNode, ERROR_DIRECTORY, 1);

    final String startOfErrorsPattern = "SEVERE: Errors:";
    String out = readFile(outPath, Charset.defaultCharset());
    assertThat(out).contains(startOfErrorsPattern);
    out = out.substring(out.indexOf(startOfErrorsPattern) + 16);

    Path expectedPath = Paths.get(TestUtils.INPUT_DIRECTORY_PREFIX)
        .resolve(ERROR_DIRECTORY)
        .resolve(expectedName + ".expected");
    String expected = readFile(expectedPath, Charset.defaultCharset());
    List<String> outList = Arrays.asList(out.split("\n"));
    Collections.sort(outList);
    List<String> expectedList = Arrays.stream(expected.split("\n"))
        .sorted()
        .filter(s -> !s.isEmpty() && !s.startsWith("//"))
        .collect(Collectors.toList());

    Assertions.assertLinesMatch(expectedList, outList);

    logger.info("ragconnect for " + expectedName + " returned:\n{}", out);
  }

}
