package org.jastadd.ragconnect.tests.singleList;

import org.jastadd.ragconnect.tests.TestUtils;
import singleList.ast.*;

import java.io.IOException;
import java.util.concurrent.TimeUnit;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Test case "list manual".
 *
 * @author rschoene - Initial contribution
 */
public class SingleListManualTest extends AbstractSingleListTest {

  private Root model;
  private MqttHandler handler;

  SingleListManualTest() {
    super("singleList");
  }

  @Override
  protected void createModel() {
    model = new Root();
    senderRoot = new SenderRoot();
    model.addSenderRoot((SenderRoot) senderRoot);

    ReceiverRoot localReceiverRoot = new ReceiverRoot();
    model.addReceiverRoot(localReceiverRoot);

    // first prepare non-wildcard lists
    for (int i = 0; i < 5; i++) {
      localReceiverRoot.addA(new A());
    }
    receiverRoot = localReceiverRoot;
    assertEquals(5, receiverRoot.getNumA());
  }

  @Override
  protected void setupReceiverAndConnectPart() throws IOException {
    model.ragconnectSetupMqttWaitUntilReady(2, TimeUnit.SECONDS);

    handler = new MqttHandler().dontSendWelcomeMessage().setHost(TestUtils.getMqttHost());
    assertTrue(handler.waitUntilReady(2, TimeUnit.SECONDS));

    // add dependencies
    ((SenderRoot) senderRoot).addInputDependencyToA1((SenderRoot) senderRoot);
    ((SenderRoot) senderRoot).addInputDependencyToA2((SenderRoot) senderRoot);
    ((SenderRoot) senderRoot).addInputDependencyToA3((SenderRoot) senderRoot);
    ((SenderRoot) senderRoot).addInputDependencyToA4((SenderRoot) senderRoot);

    data = new ReceiverData();
    handler.newConnection(TOPIC_A_WILDCARD, bytes -> data.numberOfElements += 1);
  }

  @Override
  protected void closeConnections() {
    if (handler != null) {
      handler.close();
    }
    if (model != null) {
      model.ragconnectCloseConnections();
    }
  }
}
