package org.jastadd.ragconnect.tests.singleListVariant;

import org.jastadd.ragconnect.tests.AbstractMqttTest;
import org.jastadd.ragconnect.tests.TestUtils;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.nio.file.Paths;
import java.util.List;
import java.util.function.BiConsumer;

import static java.lang.Math.abs;
import static java.util.Collections.addAll;
import static org.assertj.core.util.Lists.newArrayList;
import static org.jastadd.ragconnect.tests.TestUtils.mqttUri;
import static org.jastadd.ragconnect.tests.TestUtils.testJaddContainReferenceToJackson;
import static org.jastadd.ragconnect.tests.singleListVariant.AbstractSingleListVariantTest.IntList.list;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Base class for test cases "singleList manual" and "singleList incremental".
 *
 * @author rschoene - Initial contribution
 */
@Tag("List")
@Tag("SingleList")
public abstract class AbstractSingleListVariantTest extends AbstractMqttTest {

  public interface TestWrapperJastAddList<T> extends Iterable<T> {
    int getNumChild();
  }
  public interface TestWrapperReceiverRoot {
    TestWrapperJastAddList<? extends TestWrapperT_Empty> getT_EmptyList();
    TestWrapperJastAddList<? extends TestWrapperT_Token> getT_TokenList();
    TestWrapperJastAddList<? extends TestWrapperT_OneChild> getT_OneChildList();
    TestWrapperJastAddList<? extends TestWrapperT_OneOpt> getT_OneOptList();
    TestWrapperJastAddList<? extends TestWrapperT_OneList> getT_OneListList();
    TestWrapperJastAddList<? extends TestWrapperT_TwoChildren> getT_TwoChildrenList();
    TestWrapperJastAddList<? extends TestWrapperT_OneOfEach> getT_OneOfEachList();
    TestWrapperJastAddList<? extends TestWrapperT_Abstract> getT_AbstractList();

    TestWrapperJastAddList<? extends TestWrapperT_Empty> getMyEmptyList();

    TestWrapperJastAddList<? extends TestWrapperT_Empty> getEmptyWithAddList();
    TestWrapperJastAddList<? extends TestWrapperT_Token> getTokenWithAddList();
    TestWrapperJastAddList<? extends TestWrapperT_OneChild> getOneChildWithAddList();
    TestWrapperJastAddList<? extends TestWrapperT_OneOpt> getOneOptWithAddList();
    TestWrapperJastAddList<? extends TestWrapperT_OneList> getOneListWithAddList();
    TestWrapperJastAddList<? extends TestWrapperT_TwoChildren> getTwoChildrenWithAddList();
    TestWrapperJastAddList<? extends TestWrapperT_OneOfEach> getOneOfEachWithAddList();
    TestWrapperJastAddList<? extends TestWrapperT_Abstract> getAbstractWithAddList();

    boolean connectT_Empty(String mqttUri) throws IOException;
    boolean connectT_Token(String mqttUri) throws IOException;
    boolean connectT_OneChild(String mqttUri) throws IOException;
    boolean connectT_OneOpt(String mqttUri) throws IOException;
    boolean connectT_OneList(String mqttUri) throws IOException;
    boolean connectT_TwoChildren(String mqttUri) throws IOException;
    boolean connectT_OneOfEach(String mqttUri) throws IOException;
    boolean connectT_Abstract(String mqttUri) throws IOException;

    boolean connectMyEmpty(String mqttUri) throws IOException;

    boolean connectEmptyWithAdd(String mqttUri) throws IOException;
    boolean connectTokenWithAdd(String mqttUri) throws IOException;
    boolean connectOneChildWithAdd(String mqttUri) throws IOException;
    boolean connectOneOptWithAdd(String mqttUri) throws IOException;
    boolean connectOneListWithAdd(String mqttUri) throws IOException;
    boolean connectTwoChildrenWithAdd(String mqttUri) throws IOException;
    boolean connectOneOfEachWithAdd(String mqttUri) throws IOException;
    boolean connectAbstractWithAdd(String mqttUri) throws IOException;
  }
  @SuppressWarnings("UnusedReturnValue")
  public interface TestWrapperSenderRoot {
    boolean connectT_Empty(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectT_Token(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectT_OneChild(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectT_OneOpt(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectT_OneList(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectT_TwoChildren(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectT_OneOfEach(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectT_Abstract(String mqttUri, boolean writeCurrentValue) throws IOException;

    TestWrapperSenderRoot setInput(int input);
    TestWrapperSenderRoot setShouldSetOptAndList(boolean shouldSetOptAndList);
    TestWrapperT_Empty getT_Empty();
    TestWrapperT_OneOpt getT_OneOpt();
  }
  public interface TestWrapperNameable {
    int getID();
  }
  public interface TestWrapperOther extends TestWrapperNameable {}
  public interface TestWrapperT_Empty extends TestWrapperNameable {}
  public interface TestWrapperT_Token extends TestWrapperNameable {
    String getValue();
  }
  public interface TestWrapperT_OneChild extends TestWrapperNameable {
    TestWrapperNameable getOther();
  }
  public interface TestWrapperT_OneOpt extends TestWrapperNameable {
    boolean hasOther();
    TestWrapperNameable getOther();
  }
  public interface TestWrapperT_OneList extends TestWrapperNameable {
    int getNumOther();
    TestWrapperNameable getOther(int index);
  }
  public interface TestWrapperT_TwoChildren extends TestWrapperNameable {
    TestWrapperNameable getLeft();
    TestWrapperNameable getRight();
  }
  public interface TestWrapperT_OneOfEach extends TestWrapperNameable {
    TestWrapperNameable getFirst();
    boolean hasSecond();
    TestWrapperNameable getSecond();
    int getNumThird();
    TestWrapperNameable getThird(int index);
    String getFourth();
  }
  public interface TestWrapperT_Abstract extends TestWrapperNameable {
    String getValueAbstract();
    String getValueSub();
  }

  AbstractSingleListVariantTest(String shortName) {
    this.shortName = shortName;
  }

  protected static final String TOPIC_T_Empty = "t/Empty";
  protected static final String TOPIC_T_Token = "t/Token";
  protected static final String TOPIC_T_OneChild = "t/OneChild";
  protected static final String TOPIC_T_OneOpt = "t/OneOpt";
  protected static final String TOPIC_T_OneList = "t/OneList";
  protected static final String TOPIC_T_TwoChildren = "t/TwoChildren";
  protected static final String TOPIC_T_OneOfEach = "t/OneOfEach";
  protected static final String TOPIC_T_Abstract = "t/Abstract";
  protected static final String TOPIC_T_all = "t/#";

  protected TestWrapperSenderRoot senderRoot;
  protected TestWrapperReceiverRoot receiverRoot;
  protected ReceiverData data;

  private final String shortName;

  @Test
  public void checkJacksonReference() {
    testJaddContainReferenceToJackson(
        Paths.get("src", "test",
            "02-after-ragconnect", shortName, "RagConnect.jadd"), true);
  }

  @Override
  protected void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException, InterruptedException {
    // late model initialization
    setInput(0);
    setShouldSetOptAndList(false);

    setupReceiverAndConnectPart();

    // connect. important: first receivers, then senders. to not miss initial value.

    // receive: unnamed
    assertTrue(receiverRoot.connectT_Empty(mqttUri(TOPIC_T_Empty)));
    assertTrue(receiverRoot.connectT_Token(mqttUri(TOPIC_T_Token)));
    assertTrue(receiverRoot.connectT_OneChild(mqttUri(TOPIC_T_OneChild)));
    assertTrue(receiverRoot.connectT_OneOpt(mqttUri(TOPIC_T_OneOpt)));
    assertTrue(receiverRoot.connectT_OneList(mqttUri(TOPIC_T_OneList)));
    assertTrue(receiverRoot.connectT_TwoChildren(mqttUri(TOPIC_T_TwoChildren)));
    assertTrue(receiverRoot.connectT_OneOfEach(mqttUri(TOPIC_T_OneOfEach)));
    assertTrue(receiverRoot.connectT_Abstract(mqttUri(TOPIC_T_Abstract)));

    // receive: named
    assertTrue(receiverRoot.connectMyEmpty(mqttUri(TOPIC_T_Empty)));

    // receive: with add
    assertTrue(receiverRoot.connectEmptyWithAdd(mqttUri(TOPIC_T_Empty)));
    assertTrue(receiverRoot.connectTokenWithAdd(mqttUri(TOPIC_T_Token)));
    assertTrue(receiverRoot.connectOneChildWithAdd(mqttUri(TOPIC_T_OneChild)));
    assertTrue(receiverRoot.connectOneOptWithAdd(mqttUri(TOPIC_T_OneOpt)));
    assertTrue(receiverRoot.connectOneListWithAdd(mqttUri(TOPIC_T_OneList)));
    assertTrue(receiverRoot.connectTwoChildrenWithAdd(mqttUri(TOPIC_T_TwoChildren)));
    assertTrue(receiverRoot.connectOneOfEachWithAdd(mqttUri(TOPIC_T_OneOfEach)));
    assertTrue(receiverRoot.connectAbstractWithAdd(mqttUri(TOPIC_T_Abstract)));

    // send
    assertTrue(senderRoot.connectT_Empty(mqttUri(TOPIC_T_Empty), writeCurrentValue));
    assertTrue(senderRoot.connectT_Token(mqttUri(TOPIC_T_Token), writeCurrentValue));
    assertTrue(senderRoot.connectT_OneChild(mqttUri(TOPIC_T_OneChild), writeCurrentValue));
    assertTrue(senderRoot.connectT_OneOpt(mqttUri(TOPIC_T_OneOpt), writeCurrentValue));
    assertTrue(senderRoot.connectT_OneList(mqttUri(TOPIC_T_OneList), writeCurrentValue));
    assertTrue(senderRoot.connectT_TwoChildren(mqttUri(TOPIC_T_TwoChildren), writeCurrentValue));
    assertTrue(senderRoot.connectT_OneOfEach(mqttUri(TOPIC_T_OneOfEach), writeCurrentValue));
    assertTrue(senderRoot.connectT_Abstract(mqttUri(TOPIC_T_Abstract), writeCurrentValue));
  }

  abstract protected void setupReceiverAndConnectPart() throws IOException;

  @Override
  protected void communicateSendInitialValue() throws InterruptedException {
    // transmissions: 8 * 1 = 8
    checkTree(8, list(-0), list(0), list(-0));

    setInput(1);
    // transmissions: 8 + 8 = 16
    checkTree(16, list(-1), list(0, 1), list(-0, -1));

    setInput(1);
    // transmissions: 16
    checkTree(16, list(-1), list(0, 1), list(-0, -1));

    setShouldSetOptAndList(true);
    // transmissions: 16 + 3 = 19
    checkTree(19, list(1), list(0, 1), list(-0, -1, 1));

    setShouldSetOptAndList(true);
    // transmissions: 19
    checkTree(19, list(1), list(0, 1), list(-0, -1, 1));

    setInput(2);
    // transmissions: 19 + 8 = 27
    checkTree(27, list(2), list(0, 1, 2), list(-0, -1, 1, 2));

    setInput(5);
    // transmissions: 27 + 8 = 35
    checkTree(35, list(5), list(0, 1, 2, 5), list(-0, -1, 1, 2, 5));
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
    // transmissions: 0
    checkTree(0, list(), list(), list());

    setInput(1);
    // transmissions: 8 * 1 = 0
    checkTree(8, list(-1), list(1), list(-1));

    setInput(1);
    // transmissions: 8
    checkTree(8, list(-1), list(1), list(-1));

    setShouldSetOptAndList(true);
    // transmissions: 8 + 3 = 11
    checkTree(11, list(1), list(1), list(-1, 1));

    setShouldSetOptAndList(true);
    // transmissions: 11
    checkTree(11, list(1), list(1), list(-1, 1));

    setInput(2);
    // transmissions: 11 + 8 = 19
    checkTree(19, list(2), list(1, 2), list(-1, 1, 2));

    setInput(5);
    // transmissions: 19 + 8 = 27
    checkTree(27, list(5), list(1, 2, 5), list(-1, 1, 2, 5));
  }

  protected void setInput(int input) {
    senderRoot.setInput(input);
    assertEquals(input, senderRoot.getT_Empty().getID(), "ID value of empty");
  }

  protected void setShouldSetOptAndList(boolean shouldSetOptAndList) {
    senderRoot.setShouldSetOptAndList(shouldSetOptAndList);
    assertEquals(shouldSetOptAndList, senderRoot.getT_OneOpt().hasOther(), "opt is filled or not");
  }

  /**
   * Check against expected lists of IDs.
   * If an ID is negative, do not check Opts and Lists, but use absolute value for comparison.
   * The tests starts with ID 0 and does not use opts and lists at this point, so checking with > 0 is used.
   * @param expectedTransmissions            expected number of total transmissions
   * @param expectedList                     ids for unnamed and named endpoints without add
   * @param expectedWithAddList              ids for endpoints with add, but not those with opts and lists
   *                                         (only positive numbers can/need to be passed)
   * @param expectedWithAddListForOptAndList ids for endpoints with add and with opts and lists
   * @throws InterruptedException if interrupted in TestUtils.waitForMqtt
   */
  private void checkTree(int expectedTransmissions,
                         IntList expectedList,
                         IntList expectedWithAddList,
                         IntList expectedWithAddListForOptAndList)
      throws InterruptedException {
    TestUtils.waitForMqtt();
    assertEquals(expectedTransmissions, data.numberOfElements, "transmissions for any element");

    // check unnamed
    checkList(expectedList.toList(), receiverRoot.getT_EmptyList(), (e, n) -> {});
    checkList(expectedList.toList(), receiverRoot.getT_TokenList(),
        (e, n) -> assertEquals(Integer.toString(abs(e)), n.getValue()));
    checkList(expectedList.toList(), receiverRoot.getT_OneChildList(),
        (e, n) -> assertEquals(abs(e) + 1, n.getOther().getID()));
    checkList(expectedList.toList(), receiverRoot.getT_OneOptList(),
        (e, n) -> {
          assertEquals(e > 0, n.hasOther());
          if (n.hasOther()) {
            assertEquals(abs(e) + 1, n.getOther().getID());
          }
        });
    checkList(expectedList.toList(), receiverRoot.getT_OneListList(),
        (e, n) -> {
          assertEquals(e > 0 ? 1 : 0, n.getNumOther());
          if (n.getNumOther() > 0) {
            assertEquals(abs(e) + 1, n.getOther(0).getID());
          }
        });
    checkList(expectedList.toList(), receiverRoot.getT_TwoChildrenList(),
        (e, n) -> {
          assertEquals(abs(e) + 1, n.getLeft().getID());
          assertEquals(abs(e) + 1, n.getRight().getID());
        });
    checkList(expectedList.toList(), receiverRoot.getT_OneOfEachList(),
        (e, n) -> {
          assertEquals(abs(e) + 1, n.getFirst().getID());
          assertEquals(e > 0, n.hasSecond());
          if (n.hasSecond()) {
            assertEquals(abs(e) + 1, n.getSecond().getID());
          }
          assertEquals(e > 0 ? 1 : 0, n.getNumThird());
          if (n.getNumThird() > 0) {
            assertEquals(abs(e) + 1, n.getThird(0).getID());
          }
          assertEquals(Integer.toString(abs(e)), n.getFourth());
        });
    checkList(expectedList.toList(), receiverRoot.getT_AbstractList(),
        (e, n) -> {
          assertEquals(Integer.toString(abs(e)), n.getValueAbstract());
          assertEquals(Integer.toString(abs(e)), n.getValueSub());
        });

    // check named
    checkList(expectedList.toList(), receiverRoot.getMyEmptyList(), (e, n) -> {});

    // check with add
    checkList(expectedWithAddList.toList(), receiverRoot.getEmptyWithAddList(), (e, n) -> {});
    checkList(expectedWithAddList.toList(), receiverRoot.getTokenWithAddList(),
        (e, n) -> assertEquals(Integer.toString(abs(e)), n.getValue()));
    checkList(expectedWithAddList.toList(), receiverRoot.getOneChildWithAddList(),
        (e, n) -> assertEquals(abs(e) + 1, n.getOther().getID()));
    checkList(expectedWithAddListForOptAndList.toList(), receiverRoot.getOneOptWithAddList(),
        (e, n) -> {
          if (n.hasOther()) {
            assertEquals(abs(e) + 1, n.getOther().getID());
          }
        });
    checkList(expectedWithAddListForOptAndList.toList(), receiverRoot.getOneListWithAddList(),
        (e, n) -> {
          if (n.getNumOther() > 0) {
            assertEquals(abs(e) + 1, n.getOther(0).getID());
          }
        });
    checkList(expectedWithAddList.toList(), receiverRoot.getTwoChildrenWithAddList(),
        (e, n) -> {
          assertEquals(abs(e) + 1, n.getLeft().getID());
          assertEquals(abs(e) + 1, n.getRight().getID());
        });
    checkList(expectedWithAddListForOptAndList.toList(), receiverRoot.getOneOfEachWithAddList(),
        (e, n) -> {
          assertEquals(abs(e) + 1, n.getFirst().getID());
          if (n.hasSecond()) {
            assertEquals(abs(e) + 1, n.getSecond().getID());
          }
          if (n.getNumThird() > 0) {
            assertEquals(abs(e) + 1, n.getThird(0).getID());
          }
          assertEquals(Integer.toString(abs(e)), n.getFourth());
        });
    checkList(expectedWithAddList.toList(), receiverRoot.getAbstractWithAddList(),
        (e, n) -> {
          assertEquals(Integer.toString(abs(e)), n.getValueAbstract());
          assertEquals(Integer.toString(abs(e)), n.getValueSub());
        });
  }

  private <T extends TestWrapperNameable> void checkList(List<Integer> expectedList, TestWrapperJastAddList<T> actualList, BiConsumer<Integer, T> additionalTest) {
    assertEquals(expectedList.size(), actualList.getNumChild(), "same list size");
    int index = 0;
    for (T element : actualList) {
      assertEquals(abs(expectedList.get(index)), element.getID(), "correct ID for A");
      additionalTest.accept(expectedList.get(index), element);
      index++;
    }
  }

  protected static class ReceiverData {
    int numberOfElements = 0;
  }

  protected static class IntList {
    private final List<Integer> integers = newArrayList();
    public IntList(Integer... values) {
      addAll(integers, values);
    }

    public List<Integer> toList() {
      return integers;
    }

    public static IntList list(Integer... values) {
      return new IntList(values);
    }
  }

}
